/*************************************************************************
 Module:        sao-iis.c

 Description:   This module contains routines for communicating with 
                SAOimage or any imtool-type display server that supports 
                the IIS protocol
*************************************************************************/

#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <stdio.h>
#include "sao-iis.h"

int si_datain = -1;		/* initially no connection */
int si_dataout = -1;

short si_fb_conf = 0;		/* initially default (512x512) buffer size */
int   si_fb_width = 512;
int   si_fb_height = 512;

/*+***********************************************************************
 Function:	si_open()

 Abstract:	Open communication to SAOimage.

 Description:	This routine opens the communication channel to SAOimage.
		Currently this is either named pipes or sockets, selected
		at compile time.  The file descriptors are global.
*-***********************************************************************/

#ifdef __STDC__
int si_open(void)
#else /* K&R C */
int
si_open()
#endif /* __STDC __ */
{
    return (si_open_namedpipe ());
}

/*+***********************************************************************
 Function:	si_close()

 Abstract:	Close communication to SAOimage.

 Description:	This routine closes the communication channel to SAOimage.
*-***********************************************************************/

#ifdef __SDTC__
void
si_close(void)
#else /* K&R C */
void
si_close()
#endif /* __STDC__ */
{
    si_close_namedpipe ();
}
/*+***********************************************************************
 Function       si_set_fb_index()
 
 Abstract:      Get the index in the frame buffer config file.
 
 
 Discription:   Given the size of the image, it returns the 0-based
                index in the frame buffer config. file.  This index
                can then be passed to the imtool via the iis.t field.
                Also sets the frame buffer width and height.  *NO*
                information is sent to saoimage.  The values are saved
                for use in other routines.
*-***********************************************************************/
#ifdef __STDC__ 
int
si_set_fb_index(int n, int m, int *config_number, int *width, int *height)
#else /* K & R */
int si_set_fb_index (n, m, config_number, width, height)
    int n, m;
    int *config_number;
    int *width, *height;
#endif /* __STDC__ */
{
 
#define SZ_LINE 128
#define FBCONFIG_1      "/.imtoolrc"
    /* system wide frame buffer configurations */
#ifndef FBCONFIG_2
#define FBCONFIG_2      "/usr/local/lib/imtoolrc"
#endif
    /* two choices of system environment variable names */
#define FBCONFIG_ENV1   "imtoolrc"
#define FBCONFIG_ENV2   "IMTOOLRC"
 
    register char *ip;
    register FILE *fp;
    int nframes, Found = 0;
    char lbuf[SZ_LINE + 1], *fname, *getenv ();
 
    /* attempt to open the config file. */
    *width = -1;
    *height = -1;
    *config_number = -1;
    fp = NULL;

    if ((fname = getenv (FBCONFIG_ENV1)) || (fname = getenv (FBCONFIG_ENV2)))
      fp = fopen (fname, "r");
    if (!fp && (fname = getenv ("HOME")))
      fp = fopen (fname = lbuf, "r");
    if (!fp)
      fp = fopen (fname = FBCONFIG_2, "r");

    /* if cannot find a config file, return error */
    if (!fp) {
      fprintf (stderr, "Error: cannot find imtool configuration table\n");
      return (FAIL);
    }

    Found = 0;
    /* scan the frame buffer configuration file. */
    do {
        /* read in next line and check for eof */
        if (fgets (lbuf, SZ_LINE, fp) == NULL) {
            (void) fprintf (stderr,
                       "Error: configuration not found in %s table\n", fname);
            return (FAIL);
        }
        /* skip over leading spaces and tabs */
        for (ip = lbuf; (*ip == ' ') || (*ip == '\t'); ip++);
        /* skip comment lines and blank lines. */
        if (*ip == '\n' || *ip == '#')
            continue;
        if (!isdigit (*ip))
            continue;
        switch (sscanf (ip, "%d%d%d%d", config_number, &nframes, width, height)) {
         case 4:
             if ((*width >= n) && (*height >= m)){
               Found = 1;
             }
             break;             /* normal case */
         case 3:
             *height = *width;  /* default to square format */
             break;
         default:
             (void) fprintf (stderr, "imtool: bad config `%s'\n", ip);
             continue;
        }
        /* is this the configuration we want? */
    } while ((*width < n) || (*height < m));
    *config_number = *config_number - 1;
    fclose (fp);
    if (Found == 0)
        return (FAIL);
 
    /* remember this for use in later calls */
    si_fb_conf = *config_number;
    si_fb_width = *width;
    si_fb_height = *height;
   return (PASS);
}

/**************************************************************************/
#ifdef __STDC__
int
si_send_wcs(int w, int h, float z1, float z2)
#else /* K & R */
int si_send_wcs (w, h, z1, z2)
    int   w, h;
    float z1, z2;
#endif /* __STDC__ */
{
    IISM70 iis; 
    char   name[512], wcs_info[512];
    float  a, b, c, d, tx, ty;
    int    zt;
    int    i;
 
    strcpy (name, "SAN  DIEGO  STATE  ASTRONOMY");
    a = 1.0;
    b = 0.0;
    c = 0.0;
    d = -1.0;
    zt = 1;
    tx = ((w / 2) - (si_fb_width / 2) +1); 
    ty = ((si_fb_height / 2) + (h / 2)); 
 
    /*
     * zero out the buffer
     */
    for (i = 0; i < 512; i++)
        wcs_info[i] = '\0';
 
    /*
     * Format the string to set world coordinate parameters.
     *
     * The string is two lines:
     *
     * Image Name String\n a b c d tx ty min max xform
     *
     * where:
     *
     * X' = a*X + c*Y + tx Y' = b*X + d*Y + ty
     *
     * min is the minimum pixel value, max is the maximum pixel value.  For
     * display these pixel values will be mapped in to the range 1-200.  xform
     * defines the type of transformation to use.
     */

    sprintf (wcs_info,
             "%s\n%f %f %f %f %f %f %f %f %d",
             name, a, b, c, d, tx, ty, z1, z2, zt);
 
    /*
     * Send IIS command to set WCS
     */
    iis.subunit = WCS;
    iis.tid = IIS_WRITE | PACKED;
    iis.thingct =  - (short) ((int)(strlen(wcs_info) + 1));
    iis.z = 1;
    iis.x = 0;
    iis.y = 0;
 
    /*
     * 0-based index to frame buffer config file. First entry is 512x512, and
     * we used to assume this value throughout this library.  Now set by user.
     */
    iis.t = si_fb_conf;
 
    si_checksum (&iis);
 
    /*
     * Send the iis structure.
     */
    if (si_write ((char *) &iis, sizeof iis) == FAIL)
        return (FAIL);
 
    /*
     * Send the wcs info
     */
    if (si_write (wcs_info, -(iis.thingct)) == FAIL)
        return (FAIL);
 
    return (PASS);
}
 
/*+***********************************************************************
 Function:	si_checksum()				[PRIVATE]

 Abstract:	Compute IIS packet checksum.

 Description:	Compute IIS packet checksum.
*-***********************************************************************/

#ifdef __STDC__
static void
si_checksum(IISM70 *iis)
#else /* K&R C */
static void
si_check_sum(iis)
    IISM70 *iis;
#endif /* __STDC__ */
{

    int sum, j;
    short *p;

    iis->checksum = 0;
    for (j = 0, sum = 0, p = (short *) iis; j < 8; j++)
	sum += *p++;
    iis->checksum = 0177777 - sum;
}

/*+***********************************************************************
 Function:	si_write()				[PRIVATE]

 Abstract:	Write data to SAOimage.

 Description:	Write data to SAOimage.
*-***********************************************************************/

#ifdef __STDC__
static int si_write(char *buf, int size)
#else /* K&R C */
static int 
si_write(buf, size)
    char *buf;
    int size;
#endif /* __STDC__ */
{

    int n = 0;
    int total = 0;

    while (total < size) {
	n = write (si_dataout, buf, size - total);
	if (n <= 0) {
	    fprintf (stderr, "(si_write) can't write\n");
	    perror ("");
	    return (FAIL);
	}
	total += n;
    }
    return (PASS);
}

/*+***********************************************************************
 Function:	si_flush()				[PRIVATE]

 Abstract:	Flush data to SAOimage.

 Description:	Flush data to SAOimage.
*-***********************************************************************/

#ifdef __STDC__
static int si_flush (void)
#else /* K&R C */
static int 
si_flush()
#endif /* __STDC__ */
{
    return (si_flush_namedpipe ());
}

/*+***********************************************************************
 Function:	si_send_image()

 Abstract:	Send image to SaoImage.

 Description:	This routine sends an image to any imtool-type display
		server that supports the IIS protocol (from IRAF's fifo.c).
*-***********************************************************************/

#ifdef __STDC__
int si_send_image (int w, int h, Rawpix_t *im)
#else /* K&R */
int
si_send_image(w, h, im) 
    int w;			/* image width  */
    int h;			/* image height */
    Rawpix_t *im;		/* image data   */
#endif /* __STDC__ */
{

    IISpix_t *pixel_data;
    IISM70   iis;
    int      i, j, k, l=0, count=0;
    float    scale,z1=0, z2=0;
    int      min, max;
    int      runt_count;
    int      runt_start_row;
    int      N, row_no=0;
    int      cxfb = si_fb_width/2;	/* center x of frame buffer      */
    int      cyfb = si_fb_height/2;	/* center y of frame buffer 	 */
    int      cx   = w/2;		/* center x of image 	         */
    int      cy   = h/2;		/* center y of image 	         */
    int      x    = cxfb - cx;		/* x position of image in buffer */
    int      y    = cyfb - cy;		/* y position of image in buffer */


    /* how many complete rows can we cram into IIS max transfer of 32K? */

    N = (int) (PACKET_SIZE / (si_fb_width * sizeof(IISpix_t)));
    if (N > h)
      N = h;

    if ((pixel_data=(IISpix_t *) malloc(N*si_fb_width*sizeof (IISpix_t))) 
                                                    == (IISpix_t *) NULL) {
	fprintf (stderr, "Error allocating image buffer\n");
	return (FAIL);
    }

    /*
     * Get the minimum and maximum pixel values for later use in pixel scaling.
     */
    
    zscale((char *) im, w, h, 16, &z1, &z2, 0.25, 600, 120);

    fprintf(stderr,"z1= %f, z2=%f\n",z1,z2);

    /*
     * Send the WCS structure.
     */

    if (si_send_wcs(w, h, z1, z2) == FAIL)
      fprintf(stderr,"couldn't send wcs\n");

    /*
     * Scale the image before sending it to Display Server.
     */
   
    min   = (int) z1;
    max   = (int) z2;

    scale = (min == max) ? 0. : 200. / (z2 - z1);
    for (i=0; i < (w * h); i++)
       im[i] =   max(1, min(200,
                       (int)(scale *(float) ((int) im[i] - min)) ));
 
    /*
     * Fill in the fields of the iis structure that will not change in the
     * loop.
     */

    iis.tid = IIS_WRITE | PACKED;
    iis.subunit = MEMORY;
    iis.thingct = (short) -(N * si_fb_width  * sizeof (IISpix_t));
    iis.z = 1;
    iis.x = 0;
    iis.t = 0;


    /*
     * Send down an image
     */

    for (i = ((si_fb_height / N) - 1 ); i >= 0; i--) {

       /*
        * Fill in the Y coordinate of the iis structure.
        */
      
       iis.y = i * N; 
       si_checksum (&iis);

       /*
        * Fill the pixel array.
        * Map the image to the frame buffer.
        */

	for (j = N-1; j >= 0 ; j--) {
           if (row_no < y || row_no >= (y+h)) {
             for (k=0; k < si_fb_width; k++)
                pixel_data[j * si_fb_width + k] = 0;
           } /* if */ 
           else {
             for (k=0; k < si_fb_width; k++) 
                if ( (k < x) || (k >= (x+w)) )
                  pixel_data[j * si_fb_width + k] = 0;
                else  
                  pixel_data[j * si_fb_width + k] = (IISpix_t) im[l++];
           } /* else */ 
           row_no++;
        } /* for */  

	/*
	 * Send the iis structure.
	 */
	if (si_write ((char *) &iis, sizeof(iis)) == FAIL) {
	    free (pixel_data);
	    return (FAIL);
	}

	/*
	 * Send the pixels.
	 */
	if (si_write ((char *) pixel_data, N * si_fb_width * sizeof(IISpix_t)) 
          							    == FAIL) {
	  free (pixel_data);
	  return (FAIL);
	}
    }

    runt_start_row = N * i;

    /*
     * If necessry, send down the runt.
     */
    if ((runt_count = si_fb_height % N) != 0) {

	/*
	 * Fill in the Y coordinate of the iis structure.
	 */
	iis.y = runt_start_row;

	/*
	 * Shorten the count.
	 */
	iis.thingct = (short) -(si_fb_width * runt_count * sizeof (IISpix_t));

	si_checksum (&iis);

	/*
	 * Fill the pixel array as above.
	 * Map the image to the buffer as above.
	 */

        for (j = 0; j < runt_count; j++) {
           if (row_no < y || row_no > (y+h)) {
             for (k=0; k < si_fb_width; k++)
                pixel_data[j * si_fb_width + k] = 0;
           } /* if */
           else {
             for (k=0; k < si_fb_width; k++)
                if ( (k < x) || (k >= (x+w)) )
                  pixel_data[j * si_fb_width + k] = 0;
                else
                  pixel_data[j * si_fb_width + k] = (IISpix_t) im[l++];
           } /* else */
           row_no++;
         } /* for */

	
         /*
	 * Send the iis structure.
	 */
	if (si_write ((char *) &iis, sizeof(iis)) == FAIL) {
	    free (pixel_data);
	    return (FAIL);
	}

	/*
	 * Send the pixels.
	 */
	if (si_write ((char *) pixel_data,runt_count * si_fb_width * 
                                               sizeof (IISpix_t)) == FAIL) {
	    free (pixel_data);
	    return (FAIL);
	}
    }
    if (si_flush ()== FAIL) {
	free (pixel_data);
	return (FAIL);
    }
    free (pixel_data);
    return (PASS);
}

/*+***********************************************************************
 Function:      si_open_namedpipe()
 
 Abstract:      Open communication to SAOimage.
 
 Description:   This routine looks for the named pipes to talk to SAOimage
                and opens them for use by other routines.  The open file
                descriptors are global.
*-***********************************************************************/
#ifdef __STDC__
int si_open_namedpipe(void)
#else /* K&R C */
int
si_open_namedpipe()
#endif /* __STDC__ */
{
    char home[NAMELEN];
    char inname[NAMELEN];
    char outname[NAMELEN];
    struct stat instatbuf;
    struct stat outstatbuf;
    int gotpipes = FALSE;
    char *p, *getenv();
 
    /*
     * where is home
     */
    p = (char *) getenv ("HOME");
    if (p != NULL && *p != NULL) {
        strcpy (home, p);
    } else {
        fprintf (stderr, "(si_open_namedpipe) can't find HOME\n");
        perror ("");
        return (FAIL);
    }
 
    /*
     * first look in environment
     */
    p = (char *) getenv (INNAME);
    if (p != NULL && *p != NULL) {
        strcpy (inname, p);
        p = (char *) getenv (OUTNAME);
        if (p != NULL && *p != NULL) {
            strcpy (outname, p);
            if (stat (inname, &instatbuf) != -1)
                if (S_ISFIFO (instatbuf.st_mode))
                    if (stat (outname, &outstatbuf) != -1)
                        if (S_ISFIFO (outstatbuf.st_mode))
                            gotpipes = TRUE;
        }
    }
    if (!gotpipes) {
 
        /*
         * next look in   ~/.imt1?
         */
        strcpy (inname, home);
        strcat (inname, "/.");
        strcat (inname, I_DEVNAME);
        strcpy (outname, home);
        strcat (outname, "/.");
        strcat (outname, O_DEVNAME);
        if (stat (inname, &instatbuf) != -1)
            if (S_ISFIFO (instatbuf.st_mode))
                if (stat (outname, &outstatbuf) != -1)
                    if (S_ISFIFO (outstatbuf.st_mode))
                        gotpipes = TRUE;
    }
    if (!gotpipes) {
 
        /*
         * finally look in   /dev/imt1?
         */
        strcpy (inname, "/dev/");
        strcat (inname, I_DEVNAME);
        strcpy (outname, "/dev/");
        strcat (outname, O_DEVNAME);
        if (stat (inname, &instatbuf) != -1)
            if (S_ISFIFO (instatbuf.st_mode))
                if (stat (outname, &outstatbuf) != -1)
                    if (S_ISFIFO (outstatbuf.st_mode))
                        gotpipes = TRUE;
    }
  if (!gotpipes) {
        fprintf (stderr, "(si_open_namedpipe) can't open named pipes\n");
        perror ("");
        return (FAIL);
    }
 
    /*
     * Open the output fifo.  We have to open it ourselves first as a client to
     * get around the fifo open-no-client error.
     */
    if ((si_datain = open (outname, O_RDONLY | O_NDELAY)) != -1) {
        if ((si_dataout = open (outname, O_WRONLY | O_NDELAY)) != -1) {
            fcntl (si_dataout, F_SETFL, O_WRONLY);
        } else {
            fprintf (stderr, "(si_open_namedpipe) can't open %s\n", outname);
            perror ("");
            return (FAIL);
        }
        close (si_datain);
    } else {
        fprintf (stderr, "(si_open_namedpipe) can't open %s\n", outname);
        perror ("");
        return (FAIL);
    }
 
    /*
     * Open the input stream, a FIFO pseudodevice file used by applications to
     * send us commands and data.
     */
    if ((si_datain = open (inname, O_RDONLY | O_NDELAY)) != -1) {
        /* Clear O_NDELAY for reading. */
        fcntl (si_datain, F_SETFL, O_RDONLY);
    } else {
        fprintf (stderr, "(si_open_namedpipe) can't open %s\n", inname);
        perror ("");
        return (FAIL);
    }
 
    return (PASS);
}
 
/*+***********************************************************************
 Function:      si_close_namedpipe()
 
 Abstract:      Close communication to SAOimage.
 
 Description:   This routine closes the named pipes to talk to SAOimage.
*-***********************************************************************/
#ifdef __STDC__
void
si_close_namedpipe(void)
#else /* K&R C */
void
si_close_namedpipe()
#endif /* __STDC__ */
{
    if (si_datain != -1) {
        close (si_datain);
        si_datain = -1;
    }
    if (si_dataout != -1) {
        close (si_dataout);
        si_dataout = -1;
    }
}
 
/*+***********************************************************************
 Function:      si_flush_namedpipe()
 
 Abstract:      Flush data to SAOimage.
 
 Description:   Flush data to SAOimage.
*-***********************************************************************/
#ifdef __STDC__
int si_flush_namedpipe(void)
#else /* K&R C */
int
si_flush_namedpipe()
#endif /* __STDC__ */
{
    /*FFlush (&_iob[si_dataout]);*/     /* yes, this is gross */
    return (PASS);
}

/*+***********************************************************************
 Function:      si_send_erase()
 
 Abstract:      Send IIS command to erase
 
 Description:   Send IIS command to erase
*-***********************************************************************/
#ifdef __STDC__
int si_send_erase(void)
#else /* K&R */
int
si_send_erase()
#endif /* __STDC__ */
{
    IISM70 iis;
    short s;
 
    /*
     * Open channel if necessary.
     */
    if (si_datain == -1 || si_dataout == -1) {
        si_close ();
        if (si_open ()== FAIL)
            return (FAIL);
    }
 
    /*
     * Send IIS command to erase
     */
    iis.subunit = FEEDBACK;
    iis.tid = IIS_WRITE;
    iis.thingct = 0;
    iis.t = 0;
    iis.z = 1;
    iis.x=0;
    iis.y=0;
 
    si_checksum (&iis);
 
    /*
     * Send the iis structure.
     */
    if (si_write ((char *) &iis, sizeof iis) == FAIL)
        return (FAIL);
 
    /*
     * Send the erase command
     */
    /*s = 0;
    if (si_write ((char *) &s, 2) == FAIL)
        return (FAIL);*/
 
    return (PASS);
}
